"use strict";
// modified from https://github.com/vuejs/vue-next/blob/main/packages/compiler-sfc/src/compileScript.ts
Object.defineProperty(exports, "__esModule", { value: true });
exports.applyMacros = void 0;
const babel_1 = require("./babel");
// Special compiler macros
const DEFINE_PROPS = 'defineProps';
const DEFINE_EMITS = 'defineEmits';
const DEFINE_EXPOSE = 'defineExpose';
const WITH_DEFAULTS = 'withDefaults';
function applyMacros(nodes) {
    let hasDefinePropsCall = false;
    let hasDefineEmitCall = false;
    let propsRuntimeDecl;
    let propsRuntimeDefaults;
    let propsTypeDecl;
    let propsTypeDeclRaw;
    let emitsRuntimeDecl;
    let emitsTypeDecl;
    let emitsTypeDeclRaw;
    let exposeDecl;
    // props/emits declared via types
    const typeDeclaredProps = {};
    // record declared types for runtime props type generation
    const declaredTypes = {};
    function error(msg, 
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    node) {
        throw new Error(msg);
    }
    function processDefineProps(node) {
        if (!isCallOf(node, DEFINE_PROPS))
            return false;
        if (hasDefinePropsCall)
            error(`duplicate ${DEFINE_PROPS}() call`, node);
        hasDefinePropsCall = true;
        propsRuntimeDecl = node.arguments[0];
        // call has type parameters - infer runtime types from it
        if (node.typeParameters) {
            if (propsRuntimeDecl) {
                error(`${DEFINE_PROPS}() cannot accept both type and non-type arguments `
                    + 'at the same time. Use one or the other.', node);
            }
            propsTypeDeclRaw = node.typeParameters.params[0];
            propsTypeDecl = resolveQualifiedType(propsTypeDeclRaw, node => node.type === 'TSTypeLiteral');
            if (!propsTypeDecl) {
                error(`type argument passed to ${DEFINE_PROPS}() must be a literal type, `
                    + 'or a reference to an interface or literal type.', propsTypeDeclRaw);
            }
        }
        return true;
    }
    function processWithDefaults(node) {
        if (!isCallOf(node, WITH_DEFAULTS))
            return false;
        if (processDefineProps(node.arguments[0])) {
            if (propsRuntimeDecl) {
                error(`${WITH_DEFAULTS} can only be used with type-based `
                    + `${DEFINE_PROPS} declaration.`, node);
            }
            propsRuntimeDefaults = node.arguments[1];
        }
        else {
            error(`${WITH_DEFAULTS}' first argument must be a ${DEFINE_PROPS} call.`, node.arguments[0] || node);
        }
        return true;
    }
    function processDefineEmits(node) {
        if (!isCallOf(node, DEFINE_EMITS))
            return false;
        if (hasDefineEmitCall)
            error(`duplicate ${DEFINE_EMITS}() call`, node);
        hasDefineEmitCall = true;
        emitsRuntimeDecl = node.arguments[0];
        if (node.typeParameters) {
            if (emitsRuntimeDecl) {
                error(`${DEFINE_EMITS}() cannot accept both type and non-type arguments `
                    + 'at the same time. Use one or the other.', node);
            }
            emitsTypeDeclRaw = node.typeParameters.params[0];
            emitsTypeDecl = resolveQualifiedType(emitsTypeDeclRaw, node => node.type === 'TSFunctionType' || node.type === 'TSTypeLiteral');
            if (!emitsTypeDecl) {
                error(`type argument passed to ${DEFINE_EMITS}() must be a function type, `
                    + 'a literal type with call signatures, or a reference to the above types.', emitsTypeDeclRaw);
            }
        }
        return true;
    }
    function resolveQualifiedType(node, qualifier) {
        if (qualifier(node))
            return node;
        if (node.type === 'TSTypeReference'
            && node.typeName.type === 'Identifier') {
            const refName = node.typeName.name;
            const isQualifiedType = (node) => {
                if (node.type === 'TSInterfaceDeclaration'
                    && node.id.name === refName)
                    return node.body;
                else if (node.type === 'TSTypeAliasDeclaration'
                    && node.id.name === refName
                    && qualifier(node.typeAnnotation))
                    return node.typeAnnotation;
                else if (node.type === 'ExportNamedDeclaration' && node.declaration)
                    return isQualifiedType(node.declaration);
            };
            for (const node of nodes) {
                const qualified = isQualifiedType(node);
                if (qualified)
                    return qualified;
            }
        }
    }
    function processDefineExpose(node) {
        if (!isCallOf(node, DEFINE_EXPOSE))
            return false;
        if (exposeDecl)
            error(`duplicate ${DEFINE_EXPOSE}() call`, node);
        if (node.arguments.length !== 1)
            error(`${DEFINE_EXPOSE}() requires one argument`, node);
        exposeDecl = node.arguments[0];
        return true;
    }
    function genRuntimeProps(props) {
        const keys = Object.keys(props);
        if (!keys.length)
            return undefined;
        // check defaults. If the default object is an object literal with only
        // static properties, we can directly generate more optimzied default
        // decalrations. Otherwise we will have to fallback to runtime merging.
        const hasStaticDefaults = propsRuntimeDefaults
            && propsRuntimeDefaults.type === 'ObjectExpression'
            && propsRuntimeDefaults.properties.every(node => node.type === 'ObjectProperty' && !node.computed);
        return babel_1.t.objectExpression(Object.entries(props).map(([key, value]) => {
            const prop = hasStaticDefaults
                ? propsRuntimeDefaults.properties.find((node) => node.key.name === key)
                : undefined;
            if (prop)
                value.required = false;
            const entries = Object.entries(value).map(([key, value]) => key === 'type'
                ? babel_1.t.objectProperty(babel_1.t.identifier(key), typeof value === 'string' ? babel_1.t.identifier(value) : babel_1.t.arrayExpression(value.map((i) => babel_1.t.identifier(i))))
                : babel_1.t.objectProperty(babel_1.t.identifier(key), (0, babel_1.parseExpression)(JSON.stringify(value))));
            if (prop)
                entries.push(babel_1.t.objectProperty(babel_1.t.identifier('default'), prop.value));
            return babel_1.t.objectProperty(babel_1.t.identifier(key), babel_1.t.objectExpression(entries));
        }));
    }
    function getProps() {
        if (propsRuntimeDecl)
            return propsRuntimeDecl;
        if (propsTypeDecl) {
            extractRuntimeProps(propsTypeDecl, typeDeclaredProps, declaredTypes);
            return genRuntimeProps(typeDeclaredProps);
        }
    }
    function throwIfAwait(node) {
        if (node.type === 'AwaitExpression')
            error('top-level await is not supported in Vue 2', node);
    }
    nodes = nodes
        .map((raw) => {
        let node = raw;
        if (raw.type === 'ExpressionStatement')
            node = raw.expression;
        if (node.type === 'VariableDeclaration' && !node.declare) {
            const total = node.declarations.length;
            for (let i = 0; i < total; i++) {
                const decl = node.declarations[i];
                if (decl.init) {
                    if (processDefineEmits(decl.init))
                        decl.init = babel_1.t.memberExpression(babel_1.t.identifier('__ctx'), babel_1.t.identifier('emit'));
                    else if (processDefineProps(decl.init) || processWithDefaults(decl.init))
                        decl.init = babel_1.t.identifier('__props');
                    else
                        throwIfAwait(decl.init);
                }
            }
        }
        if (processWithDefaults(node) || processDefineEmits(node) || processDefineProps(node) || processDefineExpose(node))
            return null;
        throwIfAwait(node);
        return raw;
    })
        .filter(Boolean);
    return {
        nodes,
        props: getProps(),
        expose: exposeDecl,
    };
}
exports.applyMacros = applyMacros;
function isCallOf(node, test) {
    return !!(node
        && node.type === 'CallExpression'
        && node.callee.type === 'Identifier'
        && (typeof test === 'string'
            ? node.callee.name === test
            : test(node.callee.name)));
}
function extractRuntimeProps(node, props, declaredTypes) {
    const members = node.type === 'TSTypeLiteral' ? node.members : node.body;
    for (const m of members) {
        if ((m.type === 'TSPropertySignature' || m.type === 'TSMethodSignature')
            && m.key.type === 'Identifier') {
            let type;
            if (m.type === 'TSMethodSignature') {
                type = ['Function'];
            }
            else if (m.typeAnnotation) {
                type = inferRuntimeType(m.typeAnnotation.typeAnnotation, declaredTypes);
            }
            props[m.key.name] = {
                key: m.key.name,
                required: !m.optional,
                type: (type === null || type === void 0 ? void 0 : type.length) === 1 ? type[0] : type || 'null',
            };
        }
    }
}
function inferRuntimeType(node, declaredTypes) {
    switch (node.type) {
        case 'TSStringKeyword':
            return ['String'];
        case 'TSNumberKeyword':
            return ['Number'];
        case 'TSBooleanKeyword':
            return ['Boolean'];
        case 'TSObjectKeyword':
            return ['Object'];
        case 'TSTypeLiteral':
            // TODO (nice to have) generate runtime property validation
            return ['Object'];
        case 'TSFunctionType':
            return ['Function'];
        case 'TSArrayType':
        case 'TSTupleType':
            // TODO (nice to have) generate runtime element type/length checks
            return ['Array'];
        case 'TSLiteralType':
            switch (node.literal.type) {
                case 'StringLiteral':
                    return ['String'];
                case 'BooleanLiteral':
                    return ['Boolean'];
                case 'NumericLiteral':
                case 'BigIntLiteral':
                    return ['Number'];
                default:
                    return ['null'];
            }
        case 'TSTypeReference':
            if (node.typeName.type === 'Identifier') {
                if (declaredTypes[node.typeName.name])
                    return declaredTypes[node.typeName.name];
                switch (node.typeName.name) {
                    case 'Array':
                    case 'Function':
                    case 'Object':
                    case 'Set':
                    case 'Map':
                    case 'WeakSet':
                    case 'WeakMap':
                        return [node.typeName.name];
                    case 'Record':
                    case 'Partial':
                    case 'Readonly':
                    case 'Pick':
                    case 'Omit':
                    case 'Exclude':
                    case 'Extract':
                    case 'Required':
                    case 'InstanceType':
                        return ['Object'];
                }
            }
            return ['null'];
        case 'TSParenthesizedType':
            return inferRuntimeType(node.typeAnnotation, declaredTypes);
        case 'TSUnionType':
            return [
                ...new Set([].concat(...node.types.map(t => inferRuntimeType(t, declaredTypes)))),
            ];
        case 'TSIntersectionType':
            return ['Object'];
        default:
            return ['null']; // no runtime check
    }
}
