"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.parseSFC = exports.findReferencesForSFC = void 0;
const htmlparser2_1 = require("htmlparser2");
const compiler_core_1 = require("@vue/compiler-core");
const compiler_dom_1 = require("@vue/compiler-dom");
const shared_1 = require("@vue/shared");
const identifiers_1 = require("./identifiers");
const babel_1 = require("./babel");
const utils_1 = require("./utils");
var NodeTypes;
(function (NodeTypes) {
    NodeTypes.ROOT = 0, NodeTypes.ELEMENT = 1, NodeTypes.TEXT = 2, NodeTypes.COMMENT = 3, NodeTypes.SIMPLE_EXPRESSION = 4, NodeTypes.INTERPOLATION = 5, NodeTypes.ATTRIBUTE = 6, NodeTypes.DIRECTIVE = 7, NodeTypes.COMPOUND_EXPRESSION = 8, NodeTypes.IF = 9, NodeTypes.IF_BRANCH = 10, NodeTypes.FOR = 11, NodeTypes.TEXT_CALL = 12, NodeTypes.VNODE_CALL = 13, NodeTypes.JS_CALL_EXPRESSION = 14, NodeTypes.JS_OBJECT_EXPRESSION = 15, NodeTypes.JS_PROPERTY = 16, NodeTypes.JS_ARRAY_EXPRESSION = 17, NodeTypes.JS_FUNCTION_EXPRESSION = 18, NodeTypes.JS_CONDITIONAL_EXPRESSION = 19, NodeTypes.JS_CACHE_EXPRESSION = 20, NodeTypes.JS_BLOCK_STATEMENT = 21, NodeTypes.JS_TEMPLATE_LITERAL = 22, NodeTypes.JS_IF_STATEMENT = 23, NodeTypes.JS_ASSIGNMENT_EXPRESSION = 24, NodeTypes.JS_SEQUENCE_EXPRESSION = 25, NodeTypes.JS_RETURN_STATEMENT = 26;
})(NodeTypes || (NodeTypes = {}));
var ElementTypes;
(function (ElementTypes) {
    ElementTypes.ELEMENT = 0, ElementTypes.COMPONENT = 1, ElementTypes.SLOT = 2, ElementTypes.TEMPLATE = 3;
})(ElementTypes || (ElementTypes = {}));
const multilineCommentsRE = /\/\*\s(.|[\r\n])*?\*\//gm;
const singlelineCommentsRE = /\/\/\s.*/g;
const BUILD_IN_DIRECTIVES = new Set([
    'if',
    'else',
    'else-if',
    'for',
    'once',
    'model',
    'on',
    'bind',
    'slot',
    'slot-scope',
    'key',
    'ref',
    'text',
    'html',
    'show',
    'pre',
    'cloak',
    // 'el',
    // 'ref',
]);
function getComponents(node) {
    var _a;
    const current = node.type === NodeTypes.ELEMENT && node.tagType === ElementTypes.COMPONENT
        ? [node.tag]
        : node.type === NodeTypes.ELEMENT && node.tagType === ElementTypes.ELEMENT
            ? [node.tag]
            : [];
    const children = node.type === NodeTypes.IF
        ? node.branches
        : node.type === NodeTypes.ELEMENT
            || node.type === NodeTypes.IF_BRANCH
            || node.type === NodeTypes.FOR
            ? node.children
            : node.type === NodeTypes.TEXT
                || node.type === NodeTypes.COMMENT
                || node.type === NodeTypes.COMPOUND_EXPRESSION
                || node.type === NodeTypes.TEXT_CALL
                || node.type === NodeTypes.INTERPOLATION
                ? []
                : (_a = (0, utils_1.exhaustiveCheckReturnUndefined)(node)) !== null && _a !== void 0 ? _a : [];
    return [...current, ...children.flatMap(getComponents)];
}
function getDirectiveNames(node) {
    if (node.type === NodeTypes.ELEMENT) {
        const directives = node.props.flatMap(x => x.type === NodeTypes.DIRECTIVE ? [x.name] : []);
        return [...directives, ...node.children.flatMap(getDirectiveNames)];
    }
    else if (node.type === NodeTypes.IF) {
        return node.branches.flatMap(getDirectiveNames);
    }
    else if (node.type === NodeTypes.IF_BRANCH || node.type === NodeTypes.FOR) {
        return node.children.flatMap(getDirectiveNames);
    }
    else if (node.type === NodeTypes.INTERPOLATION
        || node.type === NodeTypes.COMPOUND_EXPRESSION
        || node.type === NodeTypes.TEXT
        || node.type === NodeTypes.COMMENT
        || node.type === NodeTypes.TEXT_CALL) {
        return [];
    }
    else {
        (0, utils_1.exhaustiveCheckReturnUndefined)(node);
        return [];
    }
}
function getFreeVariablesForText(input) {
    const identifiers = new Set();
    const inputWithPrefix = input.trimStart()[0] === '{' ? `(${input})` : input;
    const nodes = (0, babel_1.parse)(inputWithPrefix).program.body;
    nodes.forEach(node => (0, identifiers_1.getIdentifierUsages)(node, identifiers));
    return [...identifiers.values()];
}
function getFreeVariablesForPropsNode(node) {
    if (node.type === NodeTypes.DIRECTIVE) {
        const arg = node.arg === undefined ? [] : getFreeVariablesForNode(node.arg);
        const exp = node.exp === undefined ? [] : getFreeVariablesForNode(node.exp);
        return [...arg, ...exp];
    }
    return [];
}
function getFreeVariablesForNode(node) {
    var _a, _b, _c;
    if (node.type === NodeTypes.SIMPLE_EXPRESSION) {
        return node.isStatic ? [] : getFreeVariablesForText(node.content);
    }
    else if (node.type === NodeTypes.COMPOUND_EXPRESSION) {
        return node.children.flatMap(x => typeof x !== 'object' ? [] : getFreeVariablesForNode(x));
    }
    else if (node.type === NodeTypes.INTERPOLATION) {
        return getFreeVariablesForNode(node.content);
    }
    else if (node.type === NodeTypes.ELEMENT) {
        const children = node.children.flatMap(getFreeVariablesForNode);
        const directiveProps = node.props
            .flatMap(x => x.type === NodeTypes.DIRECTIVE ? [x] : []);
        const attributeProps = node.props
            .flatMap(x => x.type === NodeTypes.ATTRIBUTE ? [x] : []);
        const refNode = attributeProps.find(node => node.name === 'ref' && node.value !== undefined);
        const refIdentifier = (_a = refNode === null || refNode === void 0 ? void 0 : refNode.value) === null || _a === void 0 ? void 0 : _a.content;
        const vSlotNode = directiveProps.find(node => node.name === 'slot');
        const vSlotArgIdentifiers = (vSlotNode === null || vSlotNode === void 0 ? void 0 : vSlotNode.arg) === undefined ? [] : getFreeVariablesForNode(vSlotNode.arg);
        // TODO: Variable shadowing
        const vSlotExpVariableShadowingIdentifiers = [];
        const vForNode = directiveProps.find(node => node.name === 'for');
        const vForIdentifiers = ((_b = vForNode === null || vForNode === void 0 ? void 0 : vForNode.exp) === null || _b === void 0 ? void 0 : _b.type) === NodeTypes.SIMPLE_EXPRESSION ? getFreeVariablesForText(vForNode.exp.content.replace(/^.*\s(?:in|of)\s/, '')) : [];
        // TODO: Variable shadowing
        const vForExpVariableShadowingIdentifiers = [];
        const props = directiveProps
            .filter(({ name }) => name !== 'slot' && name !== 'for')
            .flatMap(getFreeVariablesForPropsNode);
        const shadowingIdentifiers = new Set([...vSlotExpVariableShadowingIdentifiers, ...vForExpVariableShadowingIdentifiers]);
        return [
            ...vSlotArgIdentifiers,
            refIdentifier,
            ...vForIdentifiers,
            ...([...children, ...props]).filter(x => !shadowingIdentifiers.has(x)),
        ].filter(utils_1.isNotNil);
    }
    else if (node.type === NodeTypes.FOR) {
        // If we use `baseCompiler`, we need add variable shadowing here
        // But we use `baseParse` now. So this branch will never be reached.
        // `NodeTypes.IF` and `NodeTypes.IF_BRANCH` will never be reached, also.
        // eslint-disable-next-line @typescript-eslint/no-unused-vars
        const { keyAlias, valueAlias } = node;
        return [node.source, ...node.children].flatMap(getFreeVariablesForNode);
    }
    else if (node.type === NodeTypes.IF) {
        return ((_c = node.branches) !== null && _c !== void 0 ? _c : []).flatMap(getFreeVariablesForNode);
    }
    else if (node.type === NodeTypes.IF_BRANCH) {
        return [node.condition, ...node.children]
            .filter(utils_1.isNotNil)
            .flatMap(getFreeVariablesForNode);
    }
    else if (node.type === NodeTypes.TEXT || node.type === NodeTypes.COMMENT || node.type === NodeTypes.TEXT_CALL) {
        return [];
    }
    else {
        (0, utils_1.exhaustiveCheckReturnUndefined)(node);
        return [];
    }
}
function findReferencesForSFC(code) {
    var _a, _b, _c;
    const rootNode = (0, compiler_core_1.baseParse)(code, compiler_dom_1.parserOptions);
    const templateChildNodes = rootNode.children.flatMap(node => node.type === NodeTypes.ELEMENT && node.tagType === ElementTypes.ELEMENT
        ? [node]
        : []);
    const templateNode = templateChildNodes.find(({ tag }) => tag === 'template');
    const components = (_a = templateNode === null || templateNode === void 0 ? void 0 : templateNode.children.flatMap(getComponents)) !== null && _a !== void 0 ? _a : [];
    const directives = (_b = templateNode === null || templateNode === void 0 ? void 0 : templateNode.children.flatMap(getDirectiveNames)) !== null && _b !== void 0 ? _b : [];
    const identifiers = (_c = templateNode === null || templateNode === void 0 ? void 0 : templateNode.children.flatMap(getFreeVariablesForNode)) !== null && _c !== void 0 ? _c : [];
    return {
        components,
        directives,
        identifiers,
    };
}
exports.findReferencesForSFC = findReferencesForSFC;
const htmlParserOptions = {
    xmlMode: true,
    lowerCaseTags: false,
    lowerCaseAttributeNames: false,
    recognizeSelfClosing: true,
};
function parseSFC(code, id, options) {
    var _a, _b, _c, _d;
    let templateLevel = 0;
    let inScriptSetup = false;
    let inScript = false;
    const striped = code
        .replace(multilineCommentsRE, r => ' '.repeat(r.length))
        .replace(singlelineCommentsRE, r => ' '.repeat(r.length));
    const scriptSetup = {
        start: 0,
        end: 0,
        contentStart: 0,
        contentEnd: 0,
        content: '',
        attrs: {},
        found: false,
        ast: undefined,
    };
    const script = {
        start: 0,
        end: 0,
        contentStart: 0,
        contentEnd: 0,
        content: '',
        attrs: {},
        found: false,
        ast: undefined,
    };
    let templateStart;
    let templateEnd;
    let templateLang = 'html';
    const parser = new htmlparser2_1.Parser({
        onopentag(name, attributes) {
            if (!name)
                return;
            if (name === 'template') {
                if (templateLevel === 0) {
                    templateStart = parser.endIndex + 1;
                    if (attributes.lang === 'pug')
                        templateLang = 'pug';
                }
                templateLevel += 1;
            }
            if (name === 'script') {
                if ('setup' in attributes) {
                    scriptSetup.start = parser.startIndex;
                    scriptSetup.contentStart = parser.endIndex + 1;
                    scriptSetup.attrs = attributes;
                    scriptSetup.found = true;
                    inScriptSetup = true;
                }
                else {
                    script.start = parser.startIndex;
                    script.contentStart = parser.endIndex + 1;
                    script.attrs = attributes;
                    script.found = true;
                    inScript = true;
                }
            }
        },
        onclosetag(name) {
            if (name === 'template') {
                templateLevel -= 1;
                if (templateLevel === 0 && templateStart != null)
                    templateEnd = parser.startIndex;
            }
            if (inScriptSetup && name === 'script') {
                scriptSetup.end = parser.endIndex + 1;
                scriptSetup.contentEnd = parser.startIndex;
                scriptSetup.content = code.slice(scriptSetup.contentStart, scriptSetup.contentEnd);
                inScriptSetup = false;
            }
            if (inScript && name === 'script') {
                script.end = parser.endIndex + 1;
                script.contentEnd = parser.startIndex;
                script.content = code.slice(script.contentStart, script.contentEnd);
                inScript = false;
            }
        },
    }, htmlParserOptions);
    parser.write(striped);
    parser.end();
    if (script.found
        && scriptSetup.found
        && scriptSetup.attrs.lang !== script.attrs.lang) {
        throw new SyntaxError('<script setup> language must be the same as <script>');
    }
    const parserOptions = {
        sourceType: 'module',
        plugins: [],
    };
    const lang = scriptSetup.attrs.lang || script.attrs.lang || 'js';
    if (lang === 'ts')
        parserOptions.plugins.push('typescript');
    else if (lang === 'jsx')
        parserOptions.plugins.push('jsx');
    else if (lang === 'tsx')
        parserOptions.plugins.push('typescript', 'jsx');
    else if (lang !== 'js')
        throw new SyntaxError(`Unsupported script language: ${lang}`);
    scriptSetup.ast = (0, babel_1.parse)(scriptSetup.content, parserOptions).program;
    script.ast = (0, babel_1.parse)(script.content || '', parserOptions).program;
    scriptSetup.ast
        = ((_b = (_a = options === null || options === void 0 ? void 0 : options.astTransforms) === null || _a === void 0 ? void 0 : _a.scriptSetup) === null || _b === void 0 ? void 0 : _b.call(_a, scriptSetup.ast)) || scriptSetup.ast;
    script.ast = ((_d = (_c = options === null || options === void 0 ? void 0 : options.astTransforms) === null || _c === void 0 ? void 0 : _c.script) === null || _d === void 0 ? void 0 : _d.call(_c, script.ast)) || script.ast;
    const codeOfTemplate = (() => {
        if (templateStart == null || templateEnd == null)
            return undefined;
        const templateCode = code.slice(templateStart, templateEnd);
        const html = templateLang === 'html'
            ? templateCode
            // eslint-disable-next-line @typescript-eslint/no-var-requires
            : require('pug').compile(templateCode, { filename: id })();
        return `<template>\n${html}\n</template>`;
    })();
    const a = codeOfTemplate ? findReferencesForSFC(codeOfTemplate) : undefined;
    return {
        id,
        template: {
            components: new Set(a === null || a === void 0 ? void 0 : a.components.map(utils_1.pascalize)),
            directives: new Set(a === null || a === void 0 ? void 0 : a.directives.filter(x => !BUILD_IN_DIRECTIVES.has(x)).map(shared_1.camelize)),
            identifiers: new Set(a === null || a === void 0 ? void 0 : a.identifiers),
        },
        scriptSetup,
        script,
        parserOptions,
        extraDeclarations: [],
    };
}
exports.parseSFC = parseSFC;
