"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.transformScriptSetup = void 0;
const shared_1 = require("@vue/shared");
const generator_1 = require("@babel/generator");
const utils_1 = require("@antfu/utils");
const macros_1 = require("./macros");
const identifiers_1 = require("./identifiers");
const babel_1 = require("./babel");
const utils_2 = require("./utils");
function isAsyncImport(node) {
    var _a, _b;
    if (node.type === 'VariableDeclaration') {
        const declaration = node.declarations[0];
        return ((_b = (_a = declaration === null || declaration === void 0 ? void 0 : declaration.init) === null || _a === void 0 ? void 0 : _a.callee) === null || _b === void 0 ? void 0 : _b.name) === 'defineAsyncComponent';
    }
    return false;
}
function transformScriptSetup(sfc, options) {
    var _a, _b;
    const { scriptSetup, script, template } = sfc;
    const { nodes: body, props, expose } = (0, macros_1.applyMacros)(scriptSetup.ast.body);
    const [hoisted, setupBody] = (0, utils_1.partition)(body, n => isAsyncImport(n)
        || n.type === 'ImportDeclaration'
        || n.type === 'ExportNamedDeclaration'
        || n.type.startsWith('TS'));
    // get all identifiers in `<script setup>`
    const declarations = new Set();
    (0, identifiers_1.getIdentifierDeclarations)(hoisted, declarations);
    (0, identifiers_1.getIdentifierDeclarations)(setupBody, declarations);
    const declarationArray = Array.from(declarations).filter(utils_2.isNotNil);
    // filter out identifiers that are used in `<template>`
    const returns = declarationArray
        .filter(i => template.identifiers.has(i))
        .map((i) => {
        const id = babel_1.t.identifier(i);
        return babel_1.t.objectProperty(id, id, false, true);
    });
    const components = Array.from(template.components).map(component => {
        var _a;
        return (_a = declarationArray.find(declare => declare === component)) !== null && _a !== void 0 ? _a : declarationArray.find(declare => (0, utils_2.pascalize)(declare) === component);
    }).filter(utils_2.isNotNil);
    const directiveDeclaration = Array.from(template.directives).map((directive) => {
        const identifier = declarationArray.find(declaration => declaration === `v${(0, shared_1.capitalize)(directive)}`);
        if (identifier === undefined)
            return undefined;
        return { identifier, directive };
    }).filter(utils_2.isNotNil);
    // append `<script setup>` imports to `<script>`
    const __sfc = babel_1.t.identifier('__sfc_main');
    let hasBody = false;
    const bodyNodes = script.ast.body.map((node) => {
        // replace `export default` with a temproray variable
        // `const __sfc_main = { ... }`
        if (node.type === 'ExportDefaultDeclaration') {
            hasBody = true;
            return babel_1.t.variableDeclaration('const', [
                babel_1.t.variableDeclarator(__sfc, node.declaration),
            ]);
        }
        return node;
    });
    let ast = babel_1.t.program([
        ...sfc.extraDeclarations,
        ...hoisted,
        ...bodyNodes,
    ]);
    // inject `const __sfc_main = {}` if `<script>` has default export
    if (!hasBody) {
        ast.body.push(babel_1.t.variableDeclaration('const', [
            babel_1.t.variableDeclarator(__sfc, babel_1.t.objectExpression([])),
        ]));
    }
    // inject props function
    // `__sfc_main.props = { ... }`
    if (props) {
        hasBody = true;
        ast.body.push(babel_1.t.expressionStatement(babel_1.t.assignmentExpression('=', babel_1.t.memberExpression(__sfc, babel_1.t.identifier('props')), props)));
    }
    // inject setup function
    // `__sfc_main.setup = () => {}`
    if (body.length) {
        hasBody = true;
        const returnExpr = expose
            ? babel_1.t.callExpression(babel_1.t.memberExpression(babel_1.t.identifier('Object'), babel_1.t.identifier('assign')), [babel_1.t.objectExpression(returns), expose])
            : babel_1.t.objectExpression(returns);
        const returnStatement = babel_1.t.returnStatement(returnExpr);
        ast.body.push(babel_1.t.expressionStatement(babel_1.t.assignmentExpression('=', babel_1.t.memberExpression(__sfc, babel_1.t.identifier('setup')), babel_1.t.arrowFunctionExpression([
            babel_1.t.identifier('__props'),
            babel_1.t.identifier('__ctx'),
        ], babel_1.t.blockStatement([
            ...setupBody,
            returnStatement,
        ])))));
    }
    // inject components
    // `__sfc_main.components = Object.assign({ ... }, __sfc_main.components)`
    if (components.length) {
        hasBody = true;
        const componentsObject = babel_1.t.objectExpression(components.map((i) => {
            const id = babel_1.t.identifier(i);
            return babel_1.t.objectProperty(id, id, false, true);
        }));
        ast.body.push(babel_1.t.expressionStatement(babel_1.t.assignmentExpression('=', babel_1.t.memberExpression(__sfc, babel_1.t.identifier('components')), babel_1.t.callExpression(babel_1.t.memberExpression(babel_1.t.identifier('Object'), babel_1.t.identifier('assign')), [
            componentsObject,
            babel_1.t.memberExpression(__sfc, babel_1.t.identifier('components')),
        ]))));
    }
    // inject directives
    // `__sfc_main.directives = Object.assign({ ... }, __sfc_main.directives)`
    if (directiveDeclaration.length) {
        hasBody = true;
        const directivesObject = babel_1.t.objectExpression(directiveDeclaration.map(({ directive, identifier }) => (babel_1.t.objectProperty(babel_1.t.identifier(directive), babel_1.t.identifier(identifier), false, false))));
        ast.body.push(babel_1.t.expressionStatement(babel_1.t.assignmentExpression('=', babel_1.t.memberExpression(__sfc, babel_1.t.identifier('directives')), babel_1.t.callExpression(babel_1.t.memberExpression(babel_1.t.identifier('Object'), babel_1.t.identifier('assign')), [
            directivesObject,
            babel_1.t.memberExpression(__sfc, babel_1.t.identifier('directives')),
        ]))));
    }
    if (!hasBody && !(options === null || options === void 0 ? void 0 : options.astTransforms)) {
        return {
            ast: null,
            code: '',
        };
    }
    // re-export
    // `export default __sfc_main`
    ast.body.push(babel_1.t.exportDefaultDeclaration(__sfc));
    ast = ((_b = (_a = options === null || options === void 0 ? void 0 : options.astTransforms) === null || _a === void 0 ? void 0 : _a.post) === null || _b === void 0 ? void 0 : _b.call(_a, ast, sfc)) || ast;
    return {
        ast,
        code: (0, generator_1.default)(ast).code,
    };
}
exports.transformScriptSetup = transformScriptSetup;
